using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public class Earth : MonoBehaviour {
	
	public GameObject tiny_tiny_pilgrims;
	public GameObject tiny_wolves;
	public GameObject tiny_gators;
	public GameObject tiny_sheepdog;
	public GameObject sheepdog_target;
	//public GameObject tiny_sheepdog;
	public Texture2D earth_sourcemap;
	public Texture2D level_one;
	public Texture2D level_two;
	public Texture2D level_three;
	public Texture2D level_four;
	public Texture2D level_five;
	Texture2D earth_heightmap;
	Texture2D earth_colormap;
	Texture2D earth_steepness;
	Camera main_camera;
	public Material earth_material;
	public Vector3 size = new Vector3(10,1,10);
	int earth_tex_width = 64;
	int earth_tex_height = 64;
	int mesh_width;
	int mesh_height;
	
	float water_height = 0.12f;
	float max_slope = 52.0f;
	//Vector2[] distance_to_gate;
	
	Mesh mesh;
	
	static float LinearFalloff (float distance, float inRadius) {
		return Mathf.Clamp01(1.0f - distance / inRadius);
	}

	static float GaussFalloff (float distance, float inRadius) {
		return Mathf.Clamp01 (Mathf.Pow (360.0f, -Mathf.Pow (distance / inRadius, 2.5f) - 0.01f));
	}
	
	/* From http://blog.nobel-joergensen.com/2010/10/22/spherical-coordinates-in-unity/ */
	public static void SphericalToCartesian(float radius, float polar, float elevation, out Vector3 outCart){
	        float a = radius * Mathf.Cos(elevation);
	        outCart.x = a * Mathf.Cos(polar);
	        outCart.y = radius * Mathf.Sin(elevation);
	        outCart.z = a * Mathf.Sin(polar);
	}
	public static void CartesianToSpherical(Vector3 cartCoords, out float outRadius, out float outPolar, out float outElevation){
	        if (cartCoords.x == 0)
	            cartCoords.x = Mathf.Epsilon;
	        outRadius = Mathf.Sqrt((cartCoords.x * cartCoords.x)
	                        + (cartCoords.y * cartCoords.y)
	                        + (cartCoords.z * cartCoords.z));
	        outPolar = Mathf.Atan(cartCoords.z / cartCoords.x);
	        if (cartCoords.x < 0)
	            outPolar += Mathf.PI;
	        outElevation = Mathf.Asin(cartCoords.y / outRadius);
	}

	// Use this for initialization
	void Awake () {
		if(!tiny_tiny_pilgrims) {
			Debug.LogError("Tiny pilgrims are missing from the earth!");
		}
		
		main_camera = GameObject.Find("Main Camera").GetComponent<Camera>();
		if (!main_camera) {
			Debug.LogError("Main Camera Not Found!");
		}
		
		earth_heightmap = new Texture2D(earth_tex_width, earth_tex_height, TextureFormat.RGB24, false);
		earth_colormap = new Texture2D(earth_tex_width, earth_tex_height, TextureFormat.RGB24, false);
		earth_steepness = new Texture2D(earth_tex_width, earth_tex_height, TextureFormat.RGB24, false);
		earth_heightmap.wrapMode = TextureWrapMode.Clamp;
		
		Color[] pixels = earth_heightmap.GetPixels(0);
		for(int i=0; i<pixels.Length;i++) {
			pixels[i] = new Color(0.50f, 0.50f, 0.50f);
		}
		earth_heightmap.SetPixels(pixels);
		earth_heightmap.Apply();
				
		pixels = earth_colormap.GetPixels(0);
		for(int i=0; i<pixels.Length;i++) {
			pixels[i] = new Color(0.05f, 0.05f, 0.05f);
		}
		earth_colormap.SetPixels(pixels);
		earth_colormap.Apply();
		
		pixels = earth_steepness.GetPixels(0);
		for(int i=0; i<pixels.Length;i++) {
			pixels[i] = new Color(0.0f, 0.0f, 0.0f);
		}
		earth_steepness.SetPixels(pixels);
		earth_steepness.Apply();
		
		//distance_to_gate = new Vector2[earth_tex_width*earth_tex_height];
		
		mesh_width = Mathf.Min(earth_heightmap.width, 255);
		mesh_height = Mathf.Min(earth_heightmap.height, 255);
		
		GenerateEarth();
		
		transform.GetComponent<MeshRenderer>().material.mainTexture = earth_colormap;
	}
	
	void Start () {
		/*
		//gameObject.AddComponent<Gri
		GridGraph graph = AstarPath.active.astarData.gridGraph; //First Grid Graph in the scene
		graph.width = 16;//earth_tex_width;
		graph.depth = 16;//earth_tex_height;
		graph.UpdateSizeFromWidthDepth();
		//AstarPath.OnPostScan -= PostProcessGraph;
		AstarPath.active.Scan();
		*/
		//SpawnPilgrims();
	}
		
	void RefreshHeightmap() {
		mesh = GetComponent<MeshFilter>().mesh;
		int width = mesh_width;//Mathf.Min(earth_heightmap.width, 255);
		int height = mesh_height;//Mathf.Min(earth_heightmap.height, 255);

		int y = 0;
		int x = 0;
		
		Vector3[] mesh_vertices = mesh.vertices;
		Vector3[] mesh_normals = mesh.normals;
		
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				float pixel_height = earth_heightmap.GetPixel(x, y).grayscale;
				//pixel_height = earth_heightmap.GetPixelBilinear(x, y).grayscale; // Let's see what effect this has...
				Vector3 vertex = mesh_vertices[y*width + x];
				vertex.y = pixel_height*size.y;
				mesh_vertices[y*width + x] = vertex;
			}
		}
				
		mesh.vertices = mesh_vertices;
		mesh.RecalculateNormals();
		mesh.RecalculateBounds();
		CalculateNormals();
	}
	
	void GenerateEarth() {
		
		Debug.Log("Beginning mesh generation...");
		int width = mesh_width;//Mathf.Min(earth_heightmap.width, 255);
		int height = mesh_height;//Mathf.Min(earth_heightmap.height, 255);
		
		Vector3[] vertices = new Vector3[height * width];
		Vector2[] uv = new Vector2[height * width];
		Vector4[] tangents = new Vector4[height * width];
		
		Vector2 uv_scale = new Vector2(1.0f / (width - 1), 1.0f / (height - 1));
		Vector3 size_scale = new Vector3(size.x / (width - 1), size.y, size.z / (height - 1));
		Vector3 offset = new Vector3(-size.x/2, 0f, -size.z/2f);
		
		gameObject.AddComponent<MeshFilter>();
		gameObject.AddComponent<MeshRenderer>();
		if (earth_material)
			renderer.material = earth_material;
		else
			renderer.material.color = Color.white;
		
		mesh = GetComponent<MeshFilter>().mesh;
		int y = 0;
		int x = 0;
		/*
		vertices = new Vector3[height * width];
		uv = new Vector2[height * width];
		tangents = new Vector4[height * width];
		
		uv_scale = new Vector2(1.0f / (width - 1), 1.0f / (height - 1));
		size_scale = new Vector3(size.x / (width - 1), size.y, size.z / (height - 1));
		offset = new Vector3(-size.x/2, 0f, -size.z/2f);
		*/
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				float pixel_height = earth_heightmap.GetPixel(x, y).grayscale;
				
				Vector3 vertex = new Vector3(x, pixel_height, y);
				vertices[y*width + x] = Vector3.Scale(size_scale, vertex);
				vertices[y*width + x] += offset;
				uv[y*width + x] = Vector2.Scale(new Vector2(x,y), uv_scale);
				
				Vector3 vertexL = new Vector3(x-1, earth_heightmap.GetPixel(x-1, y).grayscale, y);
				Vector3 vertexR = new Vector3(x+1, earth_heightmap.GetPixel(x+1, y).grayscale, y);
				Vector3 tan = Vector3.Scale( size_scale, vertexR - vertexL).normalized;
				tangents[y*width + x] = new Vector4(tan.x, tan.y, tan.z, -1.0f);
			}
		}
		mesh.vertices = vertices;
		mesh.uv = uv;
		
		Debug.Log("Beginning triangle generation...");
		
		int[] triangles = new int[(height - 1) * (width - 1) * 6];
		int index = 0;
		for (y=0; y<height-1; y++) {
			for (x=0; x<width-1; x++) {
				triangles[index++] = (y	* 	width) + x;
				triangles[index++] = ((y+1) * width) + x;
				triangles[index++] = (y     * width) + x + 1;

				triangles[index++] = ((y+1) * width) + x;
				triangles[index++] = ((y+1) * width) + x + 1;
				triangles[index++] = (y     * width) + x + 1;
			}
		}
		mesh.triangles = triangles;
		mesh.RecalculateNormals();
		mesh.tangents = tangents;
		Debug.Log("Mesh generation complete.");
		
		MeshCollider this_collider = gameObject.AddComponent<MeshCollider>();
	}
	
	void PushAndPullOnMap(Vector2 pixelUV, float radius, float power) {
		pixelUV.x *= earth_tex_width;
		pixelUV.y *= earth_tex_height;
		
		// Test: single-pixel full-color
		//earth_heightmap.SetPixel((int)pixelUV.x, (int)pixelUV.y, Color.white);
		
		float sqr_radius = radius * radius; 
		for(int y = 0; y < earth_tex_width; y++) {
			for(int x = 0; x < earth_tex_height; x++) {
				float sqr_mag = (new Vector2(x,y) - pixelUV).sqrMagnitude;
				//if(sqr_mag > sqr_radius)
				//	continue;
				float distance = Mathf.Sqrt(sqr_mag);
				float falloff = LinearFalloff(distance, radius);
				//Color pix = earth_heightmap.GetPixel((int)Pi
				Color pix = earth_heightmap.GetPixel(x, y);
				float one_step = 1f/16f;
				float rising = (one_step * falloff * power);
				float sinking = 1f/255f;//(power * one_step) / 10f;				
				float new_color = rising - sinking;
				float new_height_chart = pix.grayscale + new_color;
				pix += new Color(new_color, new_color, new_color);
				earth_heightmap.SetPixel(x, y, pix);
				
				if (new_height_chart < water_height) {
					earth_colormap.SetPixel(x,y, new Color(0.85f, 0.75f, 0.55f, 0.0f));
				} else {
					earth_colormap.SetPixel(x,y, new Color(0.35f, 0.85f, 0.35f, 0.0f));
				}
			}
		}
		earth_colormap.Apply();
		earth_heightmap.Apply();
		RefreshHeightmap();
	}
	
	void AddToMap(Vector2 pixelUV, float radius, float power) {
		pixelUV.x *= earth_tex_width;
		pixelUV.y *= earth_tex_height;
		
		// Test: single-pixel full-color
		//earth_heightmap.SetPixel((int)pixelUV.x, (int)pixelUV.y, Color.white);
		
		float sqr_radius = radius * radius; 
		for(int y = 0; y < earth_tex_width; y++) {
			for(int x = 0; x < earth_tex_height; x++) {
				float sqr_mag = (new Vector2(x,y) - pixelUV).sqrMagnitude;
				if(sqr_mag > sqr_radius)
					continue;
				//Debug.Log(":" + x + ":" + y + ":" + pixelUV);
				
				float distance = Mathf.Sqrt(sqr_mag);
				float falloff = GaussFalloff(distance, radius);
				//Color pix = earth_heightmap.GetPixel((int)Pi
				Color pix = earth_heightmap.GetPixel(x, y);
				float one_step = 1f/16f;
				float new_height = one_step * falloff * power;
				float new_height_chart = pix.grayscale + new_height;
				pix = new Color(Mathf.Clamp01(pix.r + new_height), Mathf.Clamp01(pix.g + new_height), Mathf.Clamp01(pix.b + new_height), Mathf.Clamp01(pix.a + 0.0f));
				
				earth_heightmap.SetPixel(x, y, pix);
				//Debug.Log(new_height + " " + water_height);
				if (new_height_chart < water_height) {
					earth_colormap.SetPixel(x,y, new Color(0.85f, 0.75f, 0.55f, 0.0f));
				} else {
					earth_colormap.SetPixel(x,y, new Color(0.35f, 0.85f, 0.35f, 0.0f));
				}
				
			}
		}
		earth_colormap.Apply();
		earth_heightmap.Apply();
		RefreshHeightmap();
	}
	
	void MoveSheepdogTarget(Vector2 uv) {
	 //("MoveSheepdogTargetToUV", uv;
		BroadcastMessage("MoveSheepdogTargetToUV", uv, SendMessageOptions.DontRequireReceiver);
	}
	
	// Update is called once per frame
	void Update () {
		/*
		if(Input.GetButton("Fire3")) {
			RaycastHit hit;
			if( Physics.Raycast(Camera.main.ScreenPointToRay(Input.mousePosition), out hit) ) {
				Renderer renderer = hit.collider.renderer;
				MeshCollider collider = hit.collider as MeshCollider;
				if(Physics.Raycast (main_camera.ScreenPointToRay(Input.mousePosition), out hit)) {
					Renderer renderer = hit.collider.renderer;
					MeshCollider collider = hit.collider as MeshCollider;
					if(renderer == transform.renderer && collider == transform.collider) {
						MoveSheepdogTarget(hit.textureCoord);
					}
				}
			}
		}*/
		
		if(Input.GetButton("Fire1") || Input.GetButton("Fire2") || Input.GetButton("Fire3")) {
			RaycastHit hit;
			if( Physics.Raycast(Camera.main.ScreenPointToRay(Input.mousePosition), out hit) ) {
				Renderer renderer = hit.collider.renderer;
				MeshCollider collider = hit.collider as MeshCollider;
			}
			float sign = 1.0f;
			if(Input.GetButton("Fire2")) {
				sign = -1.0f;
			}
			
			// Draw black pixels on texture
			if(Physics.Raycast (main_camera.ScreenPointToRay(Input.mousePosition), out hit)) {
				Renderer renderer = hit.collider.renderer;
				MeshCollider collider = hit.collider as MeshCollider;
				if(renderer == transform.renderer && collider == transform.collider) {
					if(Input.GetButton("Fire3")) {
						MoveSheepdogTarget(hit.textureCoord);
					} else {
						AddToMap(hit.textureCoord, 9.0f, sign * 20f * Time.deltaTime);
						if(Input.GetButton("Fire1")) {
							AddToMap(new Vector2(Random.value, Random.value), 14.0f, (-sign) * 40f * Time.deltaTime);
						}
					}
					//} else {
					//	PushAndPullOnMap(hit.textureCoord, 6.0f, 20f * Time.deltaTime);
					//}
				}
			}
		}
	}
	
	// Take the 2D position given and translate into 3D space, sitting on the heightmap
	public Vector3 GetPositionFromUV(Vector2 obj_coords) {
		//obj_coords.x *= earth_tex_width;
		//obj_coords.y *= earth_tex_height;
		Vector3 size_scale = new Vector3(size.x / (mesh_width - 1), size.y, size.z / (mesh_height - 1));
				
		float pixel_height = earth_heightmap.GetPixel(Mathf.RoundToInt(obj_coords.x * earth_tex_width), Mathf.RoundToInt(obj_coords.y * earth_tex_height)).grayscale;
		float pixel_height_float = earth_heightmap.GetPixelBilinear(obj_coords.x, obj_coords.y).grayscale;
		
		Vector3 return_vec = Vector3.Scale(size, new Vector3(obj_coords.x, pixel_height_float, obj_coords.y));
		Vector3 offset = new Vector3(-size.x/2, 0f, -size.z/2f);
		return_vec += offset;
		
		return return_vec;
	}
	
	// Take the 3D position and alter the y so it sits on the heightmap
	public Vector3 GetPositionFromLocal(Vector3 obj_coords) {
		//float pixel_height = earth_heightmap.GetPixel(obj_coords.x, y).grayscale;
				
		//obj_coords.y = pixel_height;
		return obj_coords; // TODO
	}
	
	public Vector2 GetNodeFromUV(Vector2 obj_coords) {
		return new Vector2(Mathf.Clamp(Mathf.RoundToInt(obj_coords.x * (earth_tex_width)), 0, earth_tex_width - 1), Mathf.Clamp(Mathf.RoundToInt(obj_coords.y * (earth_tex_height-1)), 0, earth_tex_height));
	}
	
	public float GetSlopeFromUV(Vector2 uv) {
		Vector3 cell_normal = GetNormalFromUV(uv);
		return Vector3.Angle(cell_normal, transform.up);
	}

	public bool IsCellPassableGator(Vector2 uv) {
		//Debug.Log(uv);
		
		Vector2 cell = GetNodeFromUV(uv);
		//Debug.Log(cell + " " + uv);
		float cell_height = earth_heightmap.GetPixel((int)cell.x, (int)cell.y).grayscale;
		//Debug.Log(GetAverageNormal(GetPositionFromUV(uv), 4.0f));
		//float cell_height = earth_heightmap.GetPixelBilinear(uv.x, uv.y).grayscale;
		Vector3 cell_normal = GetNormalFromUV(uv);
		//cell_normal = GetAverageNormal(GetPositionFromUV(uv), 64.0f);
		//Debug.Log(cell_normal + " " + Vector3.Angle(cell_normal, transform.up));
		if(Vector3.Angle(cell_normal, transform.up) > 20f) { // Gators don't like slopes
			return false;
		}
		return (cell_height < (water_height + 0.02f));
	}
	
	public bool IsCellPassable(Vector2 uv) {
		//Debug.Log(uv);
		
		Vector2 cell = GetNodeFromUV(uv);
		//Debug.Log(cell + " " + uv);
		float cell_height = earth_heightmap.GetPixel((int)cell.x, (int)cell.y).grayscale;
		//Debug.Log(GetAverageNormal(GetPositionFromUV(uv), 4.0f));
		//float cell_height = earth_heightmap.GetPixelBilinear(uv.x, uv.y).grayscale;
		Vector3 cell_normal = GetNormalFromUV(uv);
		//cell_normal = GetAverageNormal(GetPositionFromUV(uv), 64.0f);
		//Debug.Log(cell_normal + " " + Vector3.Angle(cell_normal, transform.up));
		if(Vector3.Angle(cell_normal, transform.up) > max_slope) {
			return false;
		}
		return (cell_height > water_height);
	}
	
	public Vector3 GetAverageNormal(Vector3 position, float radius) {
		var vertices = mesh.vertices;
		var normals = mesh.normals;
		float sqr_radius = radius * radius;
		
		Vector3 average_normal = Vector3.zero;
		
		for (int i = 0; i < vertices.Length; i++) {
			var sqr_mag = (vertices[i] - position).sqrMagnitude;
			if (sqr_mag > sqr_radius)
				continue;
			float distance = Mathf.Sqrt(sqr_mag);
			float falloff = LinearFalloff(distance, radius);
			average_normal += falloff * normals[i];
		}		
		return average_normal.normalized;
	}
	
	public Vector3 GetNormalFromUV(Vector2 uv) {
		// convert uvs to point on the grid
		Vector2 coord = GetNodeFromUV(uv);
		// Find the vertex at this grid point
		var normals = mesh.normals;
		//Debug.Log("GetNormalFromUV: " + coord + ":" + normals[(int)coord.x + ((int)coord.y * (earth_tex_width))]);
		return normals[(int)coord.x + ((int)coord.y * (earth_tex_width))];
	}
	
	public void CalculateNormals() {
		mesh.RecalculateNormals();
		var normals = mesh.normals;
		
		for (int i = 0; i < earth_tex_height; i++) {
			for(int j = 0; j < earth_tex_width; j++) {
				Vector3 normal = normals[(i*earth_tex_width)+j].normalized;
				float slope = Mathf.Clamp01((Vector3.Angle(normal, transform.up) - max_slope));
				
				if(slope > 0.0f) {
					earth_colormap.SetPixel(j, i, new Color(0.6f, 0.66f, 0.67f));
				}
				
				earth_steepness.SetPixel(j, i, new Color(slope, 0.0f, 0.0f));
			}
		}		
		earth_steepness.Apply();
		earth_colormap.Apply();
	}
	
	public void SpawnAGator() {
		GameObject new_pilgrim = (GameObject)GameObject.Instantiate(tiny_gators);
		new_pilgrim.transform.parent = transform;
		//new_pilgrim.GetComponent<TinyPilgrim>().birthplace = new Vector2(Random.Range(0,1), Random.Range(0,1));
	}
	
	public void SpawnAWolf() {
		GameObject new_pilgrim = (GameObject)GameObject.Instantiate(tiny_wolves);
		new_pilgrim.transform.parent = transform;
		//new_pilgrim.GetComponent<TinyPilgrim>().birthplace = new Vector2(Random.Range(0,1), Random.Range(0,1));
	}
	
	public void SpawnASheepdog() {
		GameObject new_pilgrim = (GameObject)GameObject.Instantiate(tiny_sheepdog);
		new_pilgrim.transform.parent = transform;
		GameObject new_pilgrim2 = (GameObject)GameObject.Instantiate(sheepdog_target);
		new_pilgrim2.transform.parent = transform;
		//new_pilgrim.GetComponent<TinyPilgrim>().birthplace = new Vector2(Random.Range(0,1), Random.Range(0,1));
	}
	
	public void SpawnAPilgrim() {
		GameObject new_pilgrim = (GameObject)GameObject.Instantiate(tiny_tiny_pilgrims);
		new_pilgrim.transform.parent = transform;
		//new_pilgrim.GetComponent<TinyPilgrim>().birthplace = new Vector2(Random.Range(0,1), Random.Range(0,1));
	}
	
	public void SpawnPilgrims() {
		Random.seed = (int)(Time.time * 2512.0f);
		int i = transform.GetComponent<LevelWatcher>().SpawnThisMany();
		do {
			SpawnAPilgrim();	
			i--;
		} while (i > 0);
		//SpawnAWolf();
		//SpawnAGator();
		//SpawnAPilgrim();
		//SpawnAPilgrim();
	}
	
	public void RestartLevel(int lev_number) {
		// Erase Pilgrims...
				
		
		// Earth cleanup...
		Color[] pixels;
		if(true) {
			switch (Random.Range(0, 4)) {
			case 1:
				earth_sourcemap = level_one;
				break;
			case 2:
				earth_sourcemap = level_two;
				break;
			case 3:
				earth_sourcemap = level_three;
				break;
			case 4:
				earth_sourcemap = level_four;
				break;
			default:
				earth_sourcemap = level_five;
				break;	
			}
			//earth_sourcemap = (Texture2D)Resources.Load("level_two", typeof(Texture2D));
			
			
			pixels = earth_sourcemap.GetPixels(0);
			//if (Random.Range(0,5) < 2) {
			//	pixels = earth_sourcemap.GetPixels(0)
			//}
		} /*else {		
			pixels = earth_heightmap.GetPixels(0);
			for(int i=0; i<pixels.Length;i++) {
				pixels[i] = new Color(0.40f, 0.40f, 0.40f);
			}
		}*/
		earth_heightmap.SetPixels(pixels);
		earth_heightmap.Apply();
		
				
		pixels = earth_colormap.GetPixels(0);
		for(int i=0; i<pixels.Length;i++) {
			pixels[i] = new Color(0.05f, 0.05f, 0.05f);
		}
		earth_colormap.SetPixels(pixels);
		earth_colormap.Apply();
		
		pixels = earth_steepness.GetPixels(0);
		for(int i=0; i<pixels.Length;i++) {
			pixels[i] = new Color(0.0f, 0.0f, 0.0f);
		}
		earth_steepness.SetPixels(pixels);
		earth_steepness.Apply();
		
		//distance_to_gate = new Vector2[earth_tex_width*earth_tex_height];
		
		mesh_width = Mathf.Min(earth_heightmap.width, 255);
		mesh_height = Mathf.Min(earth_heightmap.height, 255);
		
		RefreshHeightmap();
		AddToMap(new Vector2(0.5f, 0.5f), 64f, 0f);
		CalculateNormals();
		BroadcastMessage("Reset");
		
		if(lev_number >= 0) {
			SpawnPilgrims();
			if(lev_number > 2) {
				if(Random.Range(0,4) <= 2) {
					SpawnASheepdog();
				}
			}
			if(lev_number > 9) {
				int gators = Mathf.Clamp(Random.Range(0, Mathf.FloorToInt(lev_number/9))-3, 0, 2);
				for(int ig = 0; ig < gators; ig++) {
					SpawnAGator();
				}
			}
			if(lev_number > 5) {
				int gators = Mathf.Clamp(Random.Range(0, Mathf.FloorToInt(lev_number/5))-3, 1, 3);
				for(int ig = 0; ig < gators; ig++) {
					SpawnAWolf();
				}
			}
		}
	}
	
	public bool KillSheepNearHere(Vector2 uv, float radius) {
		// If there are any sheep within the radius, destroy them...
		var nbrhood = gameObject.GetComponentsInChildren<TinyPilgrim>();
		bool found_any = false;
		//List<TinyPilgrim> destroy_list = new List<TinyPilgrim>();
		foreach(TinyPilgrim t in nbrhood){
			Vector2 uvp = t.GetUVPosition();
			if(Vector2.Distance(uv, uvp) <= radius) {
				//destroy_list.Add(t);
				t.RemoveYourself();
				found_any = true;
			}
		}
		return found_any;
	}
	
	public Vector2 LocationOfNearestSheepdogTarget(Vector2 uv) {
		// Assume that there is only one...
		var sdt = gameObject.GetComponentInChildren<SheepdogTargScript>();
		return sdt.GetUVPosition();
	}
	
	public Vector2 LocationOfNearestSheep(Vector2 uv) {
		//SortedList<float,Vector2> neighbors = GetNearestNeighbors(uv);
		Vector2 closest = new Vector2(9999f,9999f); // TODO: Total hack. would take too long to do it right.
		
		var nbrhood = gameObject.GetComponentsInChildren<TinyPilgrim>();
		foreach(TinyPilgrim t in nbrhood){
			Vector2 uvp = t.GetUVPosition();
			if(uv == uvp)
				continue;
			float dist = (uv - uvp).magnitude;
			if((uv - uvp).sqrMagnitude < (uv - closest).sqrMagnitude) {
				closest = uvp;
			}
			
		}
		
		return closest;
	}
	
	SortedList<float, Vector2> GetNearestNeighbors(Vector2 uv) {
		var neighbors = new SortedList<float, Vector2>();
		var nbrhood = gameObject.GetComponentsInChildren<TinyPilgrim>();
		foreach(TinyPilgrim t in nbrhood){
			Vector2 uvp = t.GetUVPosition();
			if(uv == uvp)
				continue;
			float dist = (uv - uvp).magnitude;
			try {
				neighbors.Add(dist, uvp);
			} catch (System.ArgumentException) {
				
			}
		}
		return neighbors;		
	}
	
	public Vector2 GetFlocking(Vector2 uv) {
		//SortedList<float,Vector2> neighbors = GetNearestNeighbors(uv);
		TinyPilgrim[] nbrhood = gameObject.GetComponentsInChildren<TinyPilgrim>();
		Vector2 seperation = Vector2.zero;
		Vector2 cohesion = Vector2.zero;
		Vector2 alignment = Vector2.zero;
		float max_separate = 0.15f;
		float max_cohesion = 0.2f;
		int sep_count = 0;
		int coh_count = 0;
		
		
		foreach(TinyPilgrim t in nbrhood) {
			Vector2 tuv = t.GetUVPosition();
			float dist = Vector2.Distance(uv, tuv);
			if(dist > 0 && dist < max_separate) {
				Vector2 diff = (tuv - uv);
				seperation += diff.normalized / (dist);
				sep_count++;
			}
			if(dist > 0 && dist < max_cohesion) {
				cohesion += tuv;
				coh_count++;
			}			
		}
		if(coh_count > 0) {
			cohesion /= coh_count;
			cohesion = (cohesion - uv);
			//cohesion = ((cohesion - uv).normalized) / (cohesion - uv).magnitude;
		}
		if(sep_count > 0) {
			seperation /= sep_count;
		}
		cohesion = cohesion.normalized * Mathf.Clamp01(cohesion.magnitude);
		seperation = seperation.normalized * Mathf.Clamp01(seperation.magnitude);
		
		
		Vector2 steering = (cohesion + seperation).normalized * Mathf.Clamp01((cohesion + seperation).magnitude);
		if(steering.magnitude < 0.05f) {
			steering = Vector2.zero;
		}
		
		return steering;
		//seperation.Normalize();
		
		/*
		foreach(KeyValuePair<float, Vector2> i in neighbors) {
			if(i.Key > 0) {
				Vector2 diff = (i.Value - uv).normalized;
				//diff *= 1f/i.Key;
				seperation += diff;
			}
			//Vector2 new_sep = (i.Value - uv)/i.Key;// / Mathf.Sqrt(Mathf.Sqrt((uv - i.Value).magnitude));
			//seperation += new_sep;
			//Debug.Log(new_sep);
			//Debug.DrawLine(GetPositionFromUV(uv), GetPositionFromUV(new_sep));
			//cohesion += i.Value;
		}	
		
		if(seperation.sqrMagnitude > 0) {
			seperation /= neighbors.Count;
			seperation = seperation.normalized * 0.05f;
		}
		
		//cohesion /= Mathf.Max (neighbors.Count, 1);
		//Vector2 co_vec = cohesion;
		//Debug.Log((seperation) + " " + seperation.sqrMagnitude);
		*/
		//return cohesion;
		return seperation;// + ((uv - cohesion))).normalized;
	}
	
	public float GetWaterHeight() {
		return water_height;
	}
	
	protected Vector2 SteeringAvoid(Vector2 my_place_on_the_earth, Vector2 other, float radius) {
		Vector2 steering = (other - my_place_on_the_earth);
		if (steering.magnitude <= radius) {
			steering = steering.normalized *  (radius - steering.magnitude); 
		} else {
			steering = Vector2.zero;
		}
		return steering;
	}
	
	public bool IsSheepdog() {
		var sd = gameObject.GetComponentInChildren<Sheepdog>();
		if(sd) {
			return true;
		}
		return false;		
	}
	
	public Vector2 AvoidSheepdog(Vector2 uv) {
		if(IsSheepdog()) {
			return SteeringAvoid(uv, gameObject.GetComponentInChildren<Sheepdog>().GetUVPosition(),
			                     	0.4f
		    	                     // gameObject.GetComponentInChildren<Sheepdog>().GetHeight()
			                     );
		}
		//Debug.Log("No Sheepdog");
		return Vector2.zero;
	}
	
	/*
	public Vector2 GetAlignment(Vector2 uv) {
	}
	
	public Vector2 GetCohesion(Vector2 uv) {
	}
	
	public Vector2 GetSeperation(Vector2 uv) {
	}
	*/
	//public void CalulateFlowMap(out Vector2[])
	
	/*
	 * public Vector2 SeekHigherGround(Vector2 uvs) {
		Vector2 pos = GetNodeFromUV(uvs);
		Vector3 direction = DirectionHigher((int)pos.x, (int)pos.y);
		return new Vector2(direction.x, direction.y);
	}*/
	
	/*
	public Vector3 DirectionHigher(int x, int y) {
		int min_x = (x - 1) >= 0 ? (x - 1) : 0;
		int max_x = (x + 1) <= earth_tex_width ? (x + 1) : earth_tex_width;
		int min_y = (y - 1) >= 0 ? (y - 1) : 0;
		int max_y = (y + 1) <= earth_tex_height ? (x + 1) : earth_tex_height;
		float current_height = earth_heightmap.GetPixel(x, y).grayscale;
		int[] highest = new int[]{x, y};
		float highest_height = current_height;
		for(int i = min_x; i < max_x; i++) {
			for(int j = min_y; y < max_y; j++) {
				float height = earth_heightmap.GetPixel(i, j).grayscale;
				if (height > highest_height) {
					highest_height = height;
					highest[0] = i;
					highest[1] = j;
				}
			}
		}
		float steepness = highest_height - current_height;
		Vector2 direction = new Vector2(highest[0], highest[1]);
		//Vector3 polar;
		//Vector3 radius;
		//Vector3 elevation;
		//CartesianToSpherical(direction, out radius, out polar, out elevation);
		
		//Debug.Log(direction);
		return new Vector3(direction.x, direction.y, steepness);
		//return CartesianToSpherical(direction).;
	}
	*/
	
	
	//public float DirectionToGate(Vector2 uvs) {
	//}
}
