using UnityEngine;
using System.Collections;

public class LevelWatcher : MonoBehaviour {
	
	int apothesis_total = 0;
	int level_number = 0;
	int apothesis_this_level = 0;
	int number_of_pilgrims = 0;
	int pilgrims_to_spawn = 1;
	
	float last_level_start = -1.0f;
		
	enum LevelState {LS_Go, LS_Dead, LS_Wait, LS_NotStarted, LS_Between, LS_SystemStart};
	LevelState lstate = LevelWatcher.LevelState.LS_SystemStart;
	
	// Use this for initialization
	void Start () {
	
	}
	
	// Update is called once per frame
	void Update () {
		switch (lstate) {
			case LevelWatcher.LevelState.LS_Go:
			if(Input.GetKey(KeyCode.P)) {
				// Skip to next level
				if(apothesis_this_level > 0) {
					lstate = LevelWatcher.LevelState.LS_Between;
					RestartLevel();
				}
			}
			if(apothesis_this_level >= number_of_pilgrims) {
				lstate = LevelWatcher.LevelState.LS_Between;
				RestartLevel();
			}
			string score_string =  "Number of Sheep Saved: " + apothesis_total +"\nSheep left: " + number_of_pilgrims;
			if(level_number <= 1) {
				score_string += "\nSheep move towards the tower, but only if it is high enough.\nPress Q to raise the ground.";
			}
			if(level_number == 2) {
				score_string += "\nSheep can't climb the high gray cliffs.";
			}
			if(level_number >= 3 && level_number <= 4) {
				score_string += "\nPress W to tell the sheepdog where to go.\n Sheep and Wolves avoid the sheepdog.";
			}
			if(level_number >= 5 && level_number <= 6) {
				score_string += "\nWolves eat sheep, especially sheep on high mountains.";
			}
			if(level_number >= 9 && level_number <= 10) {
				score_string += "\nGators will eat sheep too.";
			}
			if(level_number >= 21 && level_number <= 23) {
				score_string += "\nPress P to go on with only the current saved sheep.";
			}
			if((level_number % 11) == 0) {
				score_string += "\nSave the sinking sheep!";
			}
			
			transform.GetComponentInChildren<GUIText>().text = score_string;
			//BroadcastMessage("ChangeScoreText", "Number of Sheep Saved: " + apothesis_total +"\nSheep left: " + number_of_pilgrims, SendMessageOptions.DontRequireReceiver);
			break;
		case LevelState.LS_Dead:
			transform.GetComponentInChildren<GUIText>().text = "Game Over!\nNumber of Sheep Saved: " + apothesis_total + "\nPress Space to Play Again";
			//BroadcastMessage("ChangeScoreText", "Game Over!\nNumber of Sheep Saved: " + apothesis_total + "Press Space to Play Again", SendMessageOptions.DontRequireReceiver);
			if(Input.GetKey("space")) {
				lstate = LevelWatcher.LevelState.LS_Wait;			
			}
			break;
		case LevelState.LS_Between:
			transform.GetComponentInChildren<GUIText>().text = "Level " + level_number + " Complete!\nTotal Sheep Saved: " + apothesis_total +"\nPress Space For Next Level";
			if(Input.GetKey("space")) {
				lstate = LevelWatcher.LevelState.LS_Go;
				RestartLevel();
			}
			break;
		case LevelState.LS_SystemStart:
				transform.GetComponent<Earth>().RestartLevel(-1);
				lstate = LevelWatcher.LevelState.LS_NotStarted;
			break;
		case LevelState.LS_NotStarted:
			transform.GetComponentInChildren<GUIText>().text = "Click in here to start";
			if(Input.GetMouseButton(0)) {
				lstate = LevelState.LS_Wait;
			}
			break;
		case LevelState.LS_Wait:
			apothesis_total = 0;
			level_number = 0;
			RestartLevel();
			break;
		}
	}
	
	public int SpawnThisMany() {
		return pilgrims_to_spawn;
	}
	
	public void RestartLevel() {
		if (apothesis_this_level == 0 && lstate != LevelWatcher.LevelState.LS_Wait) {
			// game Over
			lstate = LevelWatcher.LevelState.LS_Dead;			
			return;
		}
		if(lstate == LevelWatcher.LevelState.LS_Between) {
			return;
		}
		last_level_start = Time.time;
		pilgrims_to_spawn = Mathf.Clamp(apothesis_this_level + 1, 0, 16);
		apothesis_this_level = 0;
		level_number++;		
		transform.GetComponent<Earth>().RestartLevel(level_number);
		number_of_pilgrims = CountPilgrims();
		lstate = LevelWatcher.LevelState.LS_Go;
	}
	
	public void IncrementEaten(Transform pilgrim) {
		Destroy(pilgrim.gameObject);
		number_of_pilgrims--;// = CountPilgrims();
	}
	
	public void IncrementApothesis(Transform pilgrim) {
		transform.GetComponent<AudioSource>().Play();
		apothesis_this_level++;
		apothesis_total++;
		Destroy(pilgrim.gameObject);
	}
	
	public int CountPilgrims() {
		Object[] pilgrims = GameObject.FindObjectsOfType(typeof(TinyPilgrim));
		return pilgrims.Length;
	}
	
	public float GetTimePastBeginning() {
		if(last_level_start >= 0) {
			return (Time.time - last_level_start);
		}
		return 0.0f;
	}
}
