using UnityEngine;
using System.Collections;

public class TinyPilgrim : MonoBehaviour {
	
	protected Earth the_earth;
	public Vector2 birthplace = new Vector2(0.5f, 0.5f);
	public Vector2 my_place_on_the_earth = new Vector2();
	public Vector3 my_place_in_the_cosmos = new Vector3();
	
	float hopping_phase;
	
	bool apothesis = false;
	bool dead = false;
	
	Quaternion facing;
	
	void Awake () {

	}

	// Use this for initialization
	void Start () {
		the_earth = transform.parent.GetComponent<Earth>();
		if(!the_earth) {
			Debug.LogError("TinyPilgrim can't find the Earth!");
		}
		birthplace = new Vector2(Random.Range(0.0f,1.0f), Random.Range(0.0f,1.0f));
		my_place_on_the_earth = birthplace;
		hopping_phase = Random.Range(0.0f, 1.0f);
	}
	
	// Update is called once per frame
	void FixedUpdate () {
		if(apothesis) {
			return;
		}
		if (dead) {
			GetEaten();
			//Destroy(gameObject); // redundant
		}
		//MoveHigher();
		//MoveGateward();
		
		Vector2 steering = Vector2.zero;
		//steering += SteeringRandom();
		//steering += SteeringHigher();
		
		steering += the_earth.AvoidSheepdog(my_place_on_the_earth) * 4.0f;
				
		steering += SteeringGateward() * 1.3f;
		steering += the_earth.GetFlocking(my_place_on_the_earth)*0.8f;
		ApplySteering(steering);
		StayInBounds();
		
		// TODO: Add a movement toward local maxima
		// TODO: add flocking with other tiny pilgrims
		CleaveToTheEarth();
		transform.position = my_place_in_the_cosmos;
		Vector2 gate = transform.parent.GetComponentInChildren<DestinationPortal>().GetLocation();
		float distance_to_gate = (my_place_on_the_earth - gate).magnitude;
		if(distance_to_gate < 0.01f) {
			AchieveApothesis();
		}
		/*
		if(my_place_in_the_cosmos.y < the_earth.GetWaterHeight()) {
			GetComponent<AudioSource>().Play();
		} else {
			GetComponent<AudioSource>().Pause();
		}
		*/
		//Debug.Log(my_place_in_the_cosmos);
	}
	
	public void RemoveYourself() {
		dead = true;
	}
	
	protected void GetEaten() {
		LevelWatcher lw = transform.parent.GetComponent<LevelWatcher>();
		lw.IncrementEaten(transform);
	}
	
	protected void AchieveApothesis() {
		//transform.parent.GetComponent<LevelWatcher>().IncrementApothesis(transform);
		LevelWatcher lw = transform.parent.GetComponent<LevelWatcher>();
		lw.IncrementApothesis(transform);
		apothesis = true;
	}
	
	protected Vector2 SteeringRandom() {
		return new Vector2(Random.Range(-1.0f, 1.0f), Random.Range(-1.0f,1.0f)).normalized;
	}
	
	protected Vector2 SteeringHigher() {
		Vector3 normal = the_earth.GetNormalFromUV(my_place_on_the_earth);
		Vector2 delta = new Vector2(normal.x, normal.z);
		if(delta.magnitude < 0.1f) {
			return Vector2.zero;
		}
		return delta.normalized;
	}
	
	protected void MoveHigher() {
		float pilgrim_speed = 0.24f;
		Vector3 normal = the_earth.GetNormalFromUV(my_place_on_the_earth);
		// translate normal to 2D
		Vector2 delta = new Vector2(normal.x, normal.z).normalized * Time.deltaTime;
		if((delta * pilgrim_speed).magnitude < 0.001f) {
			return;
		}
		Vector2 next_place = my_place_on_the_earth - (delta * pilgrim_speed);
		if (the_earth.IsCellPassable(next_place) || !the_earth.IsCellPassable(my_place_on_the_earth)) {
			my_place_on_the_earth = next_place;
		}
	}
	
	protected Vector2 SteeringGateward() {
		Vector2 gate = transform.parent.GetComponentInChildren<DestinationPortal>().GetLocation();
		float max_distance = transform.parent.GetComponentInChildren<DestinationPortal>().GetHeight();
		Vector2 steering = (my_place_on_the_earth - gate);
		if (steering.magnitude <= max_distance) {
			steering = steering.normalized * (max_distance - steering.magnitude); 
		} else {
			steering = Vector2.zero;
		}
		return steering;
	}
	
	protected Vector2 SteeringAvoid(Vector2 other, float radius) {
		Vector2 steering = (other - my_place_on_the_earth);
		if (steering.magnitude <= radius) {
			steering = steering.normalized * (radius - steering.magnitude); 
		} else {
			steering = Vector2.zero;
		}
		return steering;
	}
	
	protected Vector2 SteeringSeek(Vector2 other, float radius) {
		Vector2 steering = (my_place_on_the_earth - other);
		if (steering.magnitude <= radius) {
			steering = steering.normalized * (radius - steering.magnitude); 
		} else {
			steering = Vector2.zero;
		}
		return steering;
	}
	
	protected void ApplySteering(Vector2 steering) {
		float pilgrim_speed = 0.05f;
		Vector2 delta = steering * pilgrim_speed * Time.fixedDeltaTime;
		Vector2 old_place_on_earth = my_place_on_the_earth;
		if (the_earth.IsCellPassable(my_place_on_the_earth - delta)) {
			my_place_on_the_earth -= delta;
		} else if (!the_earth.IsCellPassable(my_place_on_the_earth)){
			my_place_on_the_earth += delta;
		}
		
		if(old_place_on_earth != my_place_on_the_earth) {
			Vector2 look_target = my_place_on_the_earth + (my_place_on_the_earth - old_place_on_earth).normalized;
			transform.LookAt(the_earth.GetPositionFromUV(look_target));
		}
		
		
	}
	
	protected void MoveGateward() {
		float pilgrim_speed = 0.05f;
		Vector2 gate = transform.parent.GetComponentInChildren<DestinationPortal>().GetLocation();
		Vector2 delta = (my_place_on_the_earth - gate).normalized * Time.deltaTime;
		Vector2 next_place = my_place_on_the_earth - (delta * pilgrim_speed);
		if (the_earth.IsCellPassable(next_place)) {
			my_place_on_the_earth = next_place;
		} else if (!the_earth.IsCellPassable(my_place_on_the_earth)){
			my_place_on_the_earth += (delta * pilgrim_speed);
		}
		StayInBounds();
	}
	/*
	protected void MoveHigher() {
		Vector2 adjustment = the_earth.SeekHigherGround(my_place_on_the_earth);
		Vector2 delta = adjustment * Time.deltaTime;
		my_place_on_the_earth += delta;
	}*/
	
	// Keeps tinyPilgrim's feet on the ground
	protected void CleaveToTheEarth() {
		Vector3 pilgrim_position = the_earth.GetPositionFromUV(my_place_on_the_earth);
		float hover = 0.075f + (0.075f * Mathf.Clamp01(the_earth.GetSlopeFromUV(my_place_on_the_earth) / 90.0f));
		float max_hop = 0.3f;
		float current_hop = Mathf.Clamp01(Mathf.Abs(Mathf.Sin(Time.time + hopping_phase) / 4.0f));
		hover += (max_hop * current_hop);
		Vector3 scale_base = new Vector3(0.25f, 0.15f, 0.25f);
		float scaled_hop = (current_hop * 0.2f);
		transform.localScale = scale_base + new Vector3(-scaled_hop/4f, scaled_hop/4f, -scaled_hop/4f);
				
		pilgrim_position += new Vector3(0.0f, hover, 0.0f);
		my_place_in_the_cosmos = pilgrim_position;
		//transform.position = pilgrim_position;
		//Debug.Log(my_place_on_the_earth + " " + pilgrim_position);
	}
	
	void StayInBounds() {
		// Keep position within UV space
		my_place_on_the_earth.x = Mathf.Min(1, Mathf.Max(0, my_place_on_the_earth.x));
		my_place_on_the_earth.y = Mathf.Min(1, Mathf.Max(0, my_place_on_the_earth.y));
	}
	
	public Vector2 GetUVPosition() {
		return my_place_on_the_earth;
	}
}
