using UnityEngine;
using System.Collections;

public class Wolf : MonoBehaviour {
	
	protected Earth the_earth;
	public Vector2 birthplace = new Vector2(0.5f, 0.5f);
	public Vector2 my_place_on_the_earth = new Vector2();
	public Vector3 my_place_in_the_cosmos = new Vector3();
	
	public AudioClip eating_sheep_sfx;
	
	float hopping_phase;

	// Use this for initialization
	void Start () {
		the_earth = transform.parent.GetComponent<Earth>();
		if(!the_earth) {
			Debug.LogError("Wolf can't find the Earth!");
		}
		birthplace = new Vector2(Random.Range(0.0f,1.0f), Random.Range(0.0f,1.0f));
		my_place_on_the_earth = birthplace;
		hopping_phase = Random.Range(0.0f, 1.0f);
	
	}
	
	// Update is called once per frame
	void FixedUpdate () {
		Vector2 steering = Vector2.zero;
		float dist_to_nearest = 99f; // Magic Numbers! Yuck!
		steering += the_earth.AvoidSheepdog(my_place_on_the_earth) * 4.0f;

		steering += SteeringNearestSheep(out dist_to_nearest);
		ApplySteering(steering);
		StayInBounds();
		
		if (dist_to_nearest < 0.01f) {
			// Is it us, or is it a sheep?
			if(the_earth.KillSheepNearHere(my_place_on_the_earth, 0.01f)) {
				GetComponent<AudioSource>().PlayOneShot(eating_sheep_sfx);
			}
		}
		
		CleaveToTheEarth();
		transform.position = my_place_in_the_cosmos;	
	}
	
	protected void ApplySteering(Vector2 steering) {
		float pilgrim_speed = 0.14f;
		Vector2 delta = steering * pilgrim_speed * Time.fixedDeltaTime;
		Vector2 old_place_on_earth = my_place_on_the_earth;
		if (the_earth.IsCellPassable(my_place_on_the_earth - delta)) {
			my_place_on_the_earth -= delta;
		} else if (!the_earth.IsCellPassable(my_place_on_the_earth)){
			my_place_on_the_earth -= (delta * 0.1f);
		} else  {
			//my_place_on_the_earth -= (delta * 0.01f);
		}
		
		if(old_place_on_earth != my_place_on_the_earth) {
			Vector2 look_target = my_place_on_the_earth + (my_place_on_the_earth - old_place_on_earth).normalized;
			transform.LookAt(the_earth.GetPositionFromUV(look_target));
		}		
	}
	
	public float GetUVHeight(float y) {
		return y * 1.4f;
	}
	
	protected Vector2 SteeringNearestSheep(out float distance_to_nearest_sheep) {
		Vector2 destination = the_earth.LocationOfNearestSheep(my_place_on_the_earth);
		//Debug.Log(destination);
		float max_distance = (GetUVHeight(the_earth.GetPositionFromUV(destination).y)/2f) + 0.5f;
		Vector2 steering = (my_place_on_the_earth - destination);
		distance_to_nearest_sheep = steering.magnitude;
		if (steering.magnitude <= max_distance) {
			steering = steering.normalized * (max_distance - steering.magnitude); 
		} else {
			steering = Vector2.zero;
		}
		return steering;
	}
	
	protected Vector2 SteeringAvoid(Vector2 other, float radius) {
		Vector2 steering = (other - my_place_on_the_earth);
		if (steering.magnitude <= radius) {
			steering = steering.normalized * (radius - steering.magnitude); 
		} else {
			steering = Vector2.zero;
		}
		return steering;
	}
	
	protected Vector2 SteeringSeek(Vector2 other, float radius) {
		Vector2 steering = (my_place_on_the_earth - other);
		if (steering.magnitude <= radius) {
			steering = steering.normalized * (radius - steering.magnitude); 
		} else {
			steering = Vector2.zero;
		}
		return steering;
	}
	
	
	// Keeps tinyPilgrim's feet on the ground
	protected void CleaveToTheEarth() {
		Vector3 pilgrim_position = the_earth.GetPositionFromUV(my_place_on_the_earth);
		float hover = 0.037f + (0.075f * Mathf.Clamp01(the_earth.GetSlopeFromUV(my_place_on_the_earth) / 90.0f));
		float max_hop = 0.2f;
		float current_hop = Mathf.Clamp01(Mathf.Abs(Mathf.Sin(Time.time + hopping_phase) / 4.0f));
		hover += (max_hop * current_hop);
		Vector3 scale_base = new Vector3(0.25f, 0.15f, 0.25f);
		float scaled_hop = (current_hop * 0.2f);
		transform.localScale = scale_base + new Vector3(-scaled_hop/4f, scaled_hop/4f, -scaled_hop/4f);
				
		pilgrim_position += new Vector3(0.0f, hover, 0.0f);
		my_place_in_the_cosmos = pilgrim_position;
		//transform.position = pilgrim_position;
		//Debug.Log(my_place_on_the_earth + " " + pilgrim_position);
	}
	
	void StayInBounds() {
		// Keep position within UV space
		my_place_on_the_earth.x = Mathf.Min(1, Mathf.Max(0, my_place_on_the_earth.x));
		my_place_on_the_earth.y = Mathf.Min(1, Mathf.Max(0, my_place_on_the_earth.y));
	}
	
	public Vector2 GetUVPosition() {
		return my_place_on_the_earth;
	}
	
	public void Reset() {
		Destroy(gameObject);
	}
}